"""
(c) 2023 LLC «DACOM M»
"""

import json
import os
import threading
import yaml
from datetime import datetime
from veilrpc.states.puppet import puppet_apply
from veilrpc.utils.bazalt import CommandExecutionError, LOCAL_MODULES_PATH, LOG, LOCAL_STORAGES_PATH, BASE_NAME
from veilrpc.utils.networking import interface_exists, OvSMirror, get_output

from base.redis_dispatcher import REDIS_CLIENT

PCAP_UTILS = {
    'pcap': 'dpi-pcap-dumper',
    'merge': 'dpi-pcap-merger',
    'unloader': 'dpi-unloader'
}
PCAP_DIRS = {
    'pcap': PCAP_UTILS['pcap'].replace('-', '_'),
    'merge': PCAP_UTILS['merge'].replace('-', '_')
}
PCAP_LOCAL_CONFIGS = {
    'pcap': os.path.join(LOCAL_MODULES_PATH, f'{PCAP_DIRS["pcap"]}.yaml'),
    'merge': os.path.join(LOCAL_MODULES_PATH, f'{PCAP_DIRS["merge"]}.yaml')
}
PCAP_GLOBAL_CONFIG = {
    'pcap': os.path.join(f'/etc/{BASE_NAME}', f'{PCAP_DIRS["pcap"]}/global.yaml')
}
PCAP_GLOBAL_DEFAULT = {'user': 'root', 'space_limit': 95}


class ConfigManager:
    """
    {
        'uuid': {
            'config': dict,
            'target_node': 'bool',
            'exit_code': 'int',
            'args': 'str',
            'active': bool,
            'enable': 'bool'
        }
    }
    """

    class SERVICE:
        running = True
        stopped = False
        enable = True
        disabled = False

    class STATUS:
        service = 'SERVICE'
        active = 'ACTIVE'
        running = 'RUNNING'
        stopping = 'STOPPING'
        success = 'SUCCESS'
        failed = 'FAILED'
        partial = 'PARTIAL'
        preparing = 'PREPARING'

    def __init__(self, filename: str, config: dict):
        self.filename = filename

        self.uuid = config.pop('uuid', None)
        if not self.uuid:
            raise NotImplementedError('The configuration does not have a uuid.')
        self.data = config
        # Properties for the network status daemon
        self.target_node = False
        # Properties for the service (IaC)
        self.args = ''
        self.active = self.SERVICE.stopped
        # The service should always start when the node reboots if active == True
        # The key is left for possible situations when the service should not be started at restart
        self.enable = self.SERVICE.enable
        self.exit_code = 0

    def __getattr__(self, item):
        return self.data.get(item)

    def update(self, **kwargs) -> None:
        self.data.update(kwargs)

    def configuration(self) -> dict:
        return {
            self.uuid: {
                'config': self.data,
                'target_node': self.target_node,
                'args': self.args,
                'active': self.active,
                'enable': self.enable,
                'exit_code': self.exit_code
            }
        }

    @staticmethod
    def configurations(filename: str, default: dict, dumper: bool = True) -> dict:
        cfg_root = "dpi_pcap_dumper::instances" if dumper else "dpi_pcap_merger::instances"

        if not os.path.exists(filename):
            return default
        try:
            with open(filename, 'r') as file:
                file_data = yaml.safe_load(file)
                return file_data.get(cfg_root) if file_data else file_data
        except (OSError, yaml.YAMLError) as e:
            raise CommandExecutionError({
                'status': ConfigManager.STATUS.failed,
                'status_reason': f'Failed to read configurations. I/O error({e.errno}): {e.strerror}'
            }) from None
        except Exception as e:
            raise CommandExecutionError({
                'status': ConfigManager.STATUS.failed,
                'status_reason': f'Unexpected error: {e!s}'
            }) from None

    def read(self, dumper: bool = True) -> None:
        configurations = ConfigManager.configurations(self.filename, {}, dumper)
        if data := configurations.get(self.uuid):
            # todo: a.tsagan-manjiev:
            #  подумать над обновлением конфигурации
            self.data = data.get('config', {}) | self.data
            self.target_node = data.get('target_node', False)
            self.args = data.get('args', '')
            self.active = data.get('active', self.SERVICE.stopped)
            self.enable = data.get('enable', self.SERVICE.enable)
            self.exit_code = data.get('exit_code', 0)

    def save(self, dumper: bool = True) -> None:
        cfg_root = "dpi_pcap_dumper::instances" if dumper else "dpi_pcap_merger::instances"
        # Read all configurations
        configurations = ConfigManager.configurations(self.filename, {}, dumper)
        # Write the configuration
        try:
            with open(self.filename, 'w') as file:
                if configurations is None:
                    configurations = self.configuration()
                else:
                    configurations.update(self.configuration())
                yaml.dump({cfg_root: configurations}, file, width=256, default_flow_style=False)
        except (OSError, yaml.YAMLError) as e:
            raise CommandExecutionError({
                'status': self.STATUS.failed,
                'status_reason': f'Failed to write configuration. I/O error({e.errno}): {e.strerror}'
            }) from None
        except Exception as e:
            raise CommandExecutionError({
                'status': self.STATUS.failed,
                'status_reason': f'Unexpected error: {e!s}'
            }) from None

    def remove(self, dumper: bool = True) -> None:
        cfg_root = "dpi_pcap_dumper::instances" if dumper else "dpi_pcap_merger::instances"
        if not os.path.exists(self.filename):
            return
        configurations = ConfigManager.configurations(self.filename, {}, dumper)
        configurations.pop(self.uuid, None)
        try:
            if not configurations:
                os.remove(self.filename)
            else:
                with open(self.filename, 'w') as file:
                    yaml.dump({cfg_root: configurations}, file, width=256, default_flow_style=False)
        except (OSError, yaml.YAMLError) as e:
            raise CommandExecutionError({
                'status': self.STATUS.failed,
                'status_reason': f'Failed to remove configuration. I/O error({e.errno}): {e.strerror}'
            }) from None
        except Exception as e:
            raise CommandExecutionError({
                'status': self.STATUS.failed,
                'status_reason': f'Unexpected error: {e!s}'
            }) from None

    def set_status(self, status: str, status_reason: str = None, exit_code: int = None) -> None:
        self.data.update(status=status)
        self.active = self.SERVICE.running if status == self.STATUS.running else self.SERVICE.stopped
        if status_reason is not None:
            self.data.update(status_reason=status_reason)
        if exit_code is not None:
            self.exit_code = exit_code
        self.save()

    @property
    def response(self) -> dict:
        return {'status': self.data.get('status'), 'status_reason': self.data.get('status_reason')}


class PacketCapture:
    """
    Объект представляет собой контроллер PCAP утилиты, производящей захват пакетов.
    """

    def __init__(self, **config):
        self.uuid = config.get('uuid')
        if self.uuid is None:
            raise NotImplementedError('PacketCapture.uuid is None')
        self.config = ConfigManager(PCAP_LOCAL_CONFIGS['pcap'], config)
        self.config.read()
        if self.config.start_time:
            self.start_time = datetime.fromisoformat(self.config.start_time).astimezone()
        else:
            self.start_time = None
        if self.config.end_time:
            self.end_time = datetime.fromisoformat(self.config.end_time).astimezone()
        else:
            self.end_time = None
        self.mirror = self.__generate_mirror_names() if config else None
        self.global_config = self.get_global_config(PCAP_GLOBAL_CONFIG['pcap'])
        self.config.args = self.cmd if config else None
        self.config.target_node = self.is_target_node if config else None

    @property
    def is_target_node(self) -> bool:
        return interface_exists(**self.config.interface)

    def __generate_mirror_names(self) -> dict:
        return {
            'source_interface': self.config.interface.get('name'),
            'destination_interface': f'{self.config.interface.get("name")}_{self.uuid[:4]}',
            'name': f'mirror_{self.uuid[:4]}',
            'direction': 'both'
        }

    @staticmethod
    def get_global_config(filename):
        if not os.path.exists(filename):
            return PCAP_GLOBAL_DEFAULT
        try:
            with open(filename, 'r') as file:
                response = yaml.safe_load(file)
                return PCAP_GLOBAL_DEFAULT | response.get('pcap', {}).get('dumper', PCAP_GLOBAL_DEFAULT)
        except (OSError, yaml.YAMLError) as e:
            raise CommandExecutionError({
                'status': ConfigManager.STATUS.failed,
                'status_reason': f'Failed to read global configurations. I/O error({e.errno}): {e.strerror}'
            }) from None
        except Exception as e:
            raise CommandExecutionError({
                'status': ConfigManager.STATUS.failed,
                'status_reason': f'Unexpected error: {e!s}'
            }) from None

    @property
    def cmd(self, merge: bool = False) -> str:
        """
        Свойство объекта, подготавливающее raw команду, передаваемую утилите захвата пакетов.
        Поддерживаемые флаги см. тут: devops/spacevm-dpi/source/config.c
        """

        return ' '.join(filter(None, (
            '-u', self.global_config.get('user'),
            '-I', self.mirror.get('source_interface'),
            '-i', self.mirror.get('destination_interface'),
            '-U', self.uuid.replace('-', ''),  # for pcap C utils
            '-w', repr(self.config.data.get('filename')),
            f'-c {self.config.limits.get("packets_count", 0)}',
            f'-C {self.config.limits.get("file_size", 0)}',
            f'-t {self.config.limits.get("timeout_seconds", 0)}',
            '-m' if merge else None,
            repr(self.config.expression) if self.config.expression else None
        )))

    def log(self, response: str = None) -> None:
        LOG.info(f"{PCAP_UTILS['pcap']} {self.uuid}: {response or self.config.response}")

    def unload_ebpf(self):
        cmd = f"{PCAP_UTILS['unloader']} {self.mirror.get('destination_interface')}"
        ret = get_output(cmd, suppress_err=True)
        self.log(f"Unload eBPF: {dict(status=ret['result'], status_reason=ret['comment'])}")

    def service_status(self, service: str) -> dict:
        actives = {'running': self.config.STATUS.running,
                   'active': self.config.STATUS.running,
                   'activating': self.config.STATUS.running,
                   'inactive': self.config.STATUS.success,
                   'exited': self.config.STATUS.success,
                   'dead': self.config.STATUS.success,
                   'failed': self.config.STATUS.failed}
        cmd = f"systemctl show {service}-{self.uuid}.service -p ExecMainCode -p ActiveState -p SubState"
        ret = get_output(cmd)
        states = dict(map(lambda s: (s.split("=")[0], s.split("=")[1]), ret['comment'].strip().splitlines()))
        is_active = states.get('ActiveState')
        if is_active == 'active' and states.get('SubState') != 'running':
            is_active = states.get('SubState')
        self.log(f"Service {service}-{self.uuid} status: "
                 f"{dict(status=is_active, exit_code=int(states.get('ExecMainCode', 0)))}")
        return {
            'status': status if (status := actives.get(is_active, actives['inactive'])) else self.config.STATUS.service,
            'exit_code': int(states.get('ExecMainCode', 0))
        }

    @property
    def is_partial(self) -> bool:
        path, filename = os.path.split(self.config.data.get('filename'))
        return os.path.exists(path) and any(filter(lambda x: x.startswith(f"{filename}.part"), os.listdir(path)))

    def parts_filesize(self) -> int:
        path, filename = os.path.split(self.config.filename)
        size = 0
        for file in os.listdir(path):
            if file.startswith(f"{filename}.part_"):
                size += os.path.getsize(os.path.join(path, file))
        return int(size / (1024 * 1024))

    def get_disk_usage(self, redis_client=REDIS_CLIENT) -> tuple:
        """
        Получает размер файловой системы, в которой существует папки и количество свободного места в ней.
        """

        storage_status = json.loads(redis_client.get('storage_status_local').decode('utf-8'))
        filepath, filename = os.path.split(self.config.data.get('filename'))
        space = {'free': 0, 'size': 0}
        for path in storage_status.get('datapools_paths', {}):
            if path in filepath:
                if path in storage_status.get('local_storages', {}).get(LOCAL_STORAGES_PATH, {}).keys():
                    space.update(storage_status.get('local_storages', {})
                                 .get(LOCAL_STORAGES_PATH, {})
                                 .get(path, {})
                                 .get('space', {}))
                elif path in storage_status.get('cluster_fs_datapools', {}).keys():
                    space.update(storage_status.get('cluster_fs_datapools', {}).get(path).get('space', {}))
                else:
                    for shared in storage_status.get('shared_storages', {}).values():
                        for volume in shared.values():
                            if any(x == path for x in volume.keys()):
                                space.update(volume.get('space', {}))
                                break
                    if not space.get('size'):
                        for zfs in storage_status.get('zfs_pools', {}).values():
                            if zfs.get('datapool', {}).get('path') in path:
                                space.update(zfs.get('datapool', {}).get('space', {}))
                                break

        return tuple(space.values())

    @property
    def is_disk_free_avail(self) -> bool:
        free, size = self.get_disk_usage()
        if size == 0:
            return False
        return round((size - free) / size * 100, 1) < self.global_config.get('space_limit')

    @property
    def is_merge_avail(self) -> bool:
        free, size = self.get_disk_usage()
        return self.parts_filesize() < free and self.is_disk_free_avail

    def start(self) -> None:
        if self.mirror is None:
            raise CommandExecutionError({'status': self.config.STATUS.failed,
                                         'status_reason': 'Packet Capturing Mirror is not configured.'})
        mirror = OvSMirror(**self.mirror)
        if self.end_time and not (condition := datetime.now(tz=self.end_time.tzinfo) < self.end_time):
            self.config.set_status(self.config.STATUS.failed,
                                   "Packet Capturing is not running, the time interval has expired.")
        else:
            condition = datetime.now(tz=self.start_time.tzinfo) > self.start_time
        if self.is_target_node and condition:
            # target node and time has come
            try:
                created = mirror.create()
            except CommandExecutionError as e:
                self.config.set_status(self.config.STATUS.failed,
                                       f'Packet Capturing is started failed. {e!r}')
                raise CommandExecutionError(self.config.response)
            else:
                if created:
                    # create mirroring
                    self.config.set_status(self.config.STATUS.running, "Packet Capturing is started successfully.")
                    ret = puppet_apply(modules=[PCAP_DIRS['pcap']])
                    if not ret['result']:
                        self.unload_ebpf()
                        mirror.remove()
                        self.config.set_status(self.config.STATUS.failed, ret['comment'], 1)
                        raise CommandExecutionError(self.config.response)
                else:
                    self.config.set_status(self.config.STATUS.failed,
                                           'Packet Capturing is started failed. Can not created mirroring interface...')
                    raise CommandExecutionError(self.config.response)
        elif not self.is_target_node or not condition:
            # not target node or time has not come
            self.config.set_status(self.config.STATUS.active, "Packet Capturing start condition is not satisfied.")
        self.log()

    def stop(self) -> None:
        if self.config.status in (self.config.STATUS.active, self.config.STATUS.running):
            if self.is_target_node:
                self.config.set_status(self.config.STATUS.stopping, "Packet Capturing is stopping...")
                ret = puppet_apply(modules=[PCAP_DIRS['pcap']])
                if not ret['result']:
                    self.config.set_status(self.config.STATUS.failed, ret['comment'], 1)
                    raise CommandExecutionError(self.config.response)
            else:
                self.config.set_status(self.config.STATUS.service, "Packet Capturing is stopped successfully.")
        self.log()

    def merger_start(self) -> None:
        if self.is_partial:
            # Packet Capturing partial files found
            if self.is_merge_avail:
                # Disk free space is enough
                data = {
                    'uuid': self.uuid,
                    'filename': self.config.data.get('filename'),
                    'parts': [],
                }
                path, filename = os.path.split(data.get('filename'))
                data['parts'].extend(filter(lambda x: x.startswith(f"{filename}.part_"), os.listdir(path)))
                config = ConfigManager(PCAP_LOCAL_CONFIGS['merge'], data)
                config.update(status=config.STATUS.preparing)
                config.args = (f"-w {data.get('filename')} "
                               f"{' '.join(map(lambda x: os.path.join(path, x), data.get('parts')))}")
                config.active = config.SERVICE.running
                config.enable = True
                config.save(dumper=False)
                ret = puppet_apply(modules=[PCAP_DIRS['merge']])
                if not ret['result']:
                    self.config.partial(ret['comment'])
                    raise CommandExecutionError(self.config.response)
                self.config.set_status(self.config.STATUS.preparing)
                self.log()
            else:
                self.config.set_status(
                    self.config.STATUS.partial,
                    'Packet Capturing merging: disk space is not enough.' if self.config.exit_code == 0 else None)
                raise CommandExecutionError(self.config.response)
        else:
            self.config.set_status(
                self.config.STATUS.service,
                'Packet Capturing is not partial: partial files not found...' if self.config.exit_code == 0 else None)
            raise CommandExecutionError(self.config.response)

    def merger_stop(self):
        if self.config.status == self.config.STATUS.preparing:
            status = self.service_status(PCAP_DIRS['merge'])
            if status == self.config.STATUS.running:
                config = ConfigManager(PCAP_LOCAL_CONFIGS['merge'], {'uuid': self.uuid})
                config.read(dumper=False)
                config.active = config.SERVICE.stopped
                self.config.set_status(self.config.STATUS.partial, "Packet Capturing merging is stopping...")
                ret = puppet_apply(modules=[PCAP_DIRS['merge']])
                if not ret['result']:
                    self.config.set_status(self.config.STATUS.partial, ret['comment'])
                    raise CommandExecutionError(self.config.response)

    def execute(self, action: str) -> dict:
        LOG.info(f"PCAP {self.uuid}: Execute {action} with config: {self.config.data} and "
                 f"star_time: {self.start_time}, end_time: {self.end_time}")
        if action == 'start' and self.config and self.config.status in (self.config.STATUS.service,
                                                                        self.config.STATUS.failed,
                                                                        self.config.STATUS.success):
            self.config.set_status(self.config.STATUS.active)
            self.start()
        elif action == 'stop':
            self.config.read()
            self.start_time = datetime.fromisoformat(self.config.start_time)
            if self.end_time:
                if (self.end_time <= datetime.now(tz=self.start_time.tzinfo) and
                        self.config.status in (self.config.STATUS.running, self.config.STATUS.active)):
                    # Stop Packet Capturing
                    self.stop()
                else:
                    # Update Packet Capturing data
                    self.config.save()
            else:
                # Stop Packet Capturing
                self.stop()
        elif action == 'merge':
            self.config.read()
            if self.config.status == self.config.STATUS.partial:
                self.merger_start()
            else:
                raise CommandExecutionError({'status': self.config.status,
                                             'status_reason': 'Packet Capturing status is not partial...'})
        elif action == 'clear':
            if self.config.status == self.config.STATUS.partial:
                path, filename = os.path.split(self.config.data.get('filename'))
                parts = list(filter(lambda x: x.startswith(f"{filename}.part_"), os.listdir(path)))
                if parts:
                    for part in parts:
                        threading.Thread(target=lambda: os.remove(os.path.join(path, part))).start()
                    self.config.set_status(self.config.STATUS.service, "Packet Capturing partial files are removing...")
                else:
                    self.config.set_status(self.config.STATUS.service, "Packet Capturing partial files not found...")
            else:
                raise CommandExecutionError({'status': self.config.status,
                                             'status_reason': 'Packet Capturing status is not partial...'})
            self.log()
        elif action == 'update':
            configurations = ConfigManager.configurations(PCAP_LOCAL_CONFIGS['pcap'], {})
            if self.uuid in configurations.keys():
                # Configuration exists, checking status it.
                config = ConfigManager(PCAP_LOCAL_CONFIGS['pcap'], {'uuid': self.uuid})
                if config and config.status == self.config.STATUS.active:
                    self.config.save()
            else:
                # Configuration not exists, creating it.
                self.config.save()
        elif action == "get":
            self.config.read()
            return self.config.data
        elif action == "remove":
            self.config.read()
            status = self.service_status(PCAP_DIRS['pcap'])
            if status in (self.config.STATUS.success, self.config.STATUS.service, self.config.STATUS.failed):
                self.config.remove()
            self.log(f"Packet Capturing configuration with status {status} is removed.")

        return self.config.response
