"""
(c) 2024 LLC «DACOM M»
"""

import os
import fcntl
import logging
from time import sleep
from datetime import datetime as dt
from veilrpc.utils.bazalt import ListVars


LOG = logging.getLogger('bazalt')
METRICS_TMPFS = '/var/www/metrics/v1/'


class MetricTypeNames(ListVars):
    storage_status = 'storage_status'


class MetricsWriter:
    def __init__(self, metric_type, tmpfs_dir=METRICS_TMPFS):
        if metric_type in MetricTypeNames.list():
            self.tmpfs_dir = tmpfs_dir
            self.prefix = metric_type
            self.symlink_path = os.path.join(self.tmpfs_dir, f'{self.prefix}.html')
            self._filename = None
        else:
            LOG.error(f'Metrics type {metric_type} is invalid, use one of: {" ,".join(MetricTypeNames.list())}')

    @property
    def _suffix(self):
        return dt.now().strftime('%Y%m%d_%H%M%S')

    def write_data(self, data):
        old_file = self._filename if self._filename else None
        self._filename = os.path.join(self.tmpfs_dir, f"{self.prefix}.html_{self._suffix}")
        if not os.path.isdir(self.tmpfs_dir):
            LOG.error(f'Failed to write metrics {self.prefix}, {self.tmpfs_dir} not found.')
            return
        list_files = os.listdir(self.tmpfs_dir)
        for f in list_files:
            f_path = os.path.join(self.tmpfs_dir, f)
            if f_path not in [old_file, self._filename, self.symlink_path]:
                os.remove(f_path)
        try:
            with open(self._filename, 'wb') as f:
                f.write(data)
        except Exception as e:
            LOG.error(f'Failed to write data metrics {self.prefix}: {str(e)}')
            return
        os.chmod(self._filename, 0o644)
        if not old_file and not os.path.islink(self.symlink_path):
            os.symlink(self._filename, self.symlink_path)
        elif old_file and not os.path.islink(self.symlink_path):
            os.symlink(self._filename, self.symlink_path)
            if os.path.exists(old_file):
                os.remove(old_file)
        elif old_file and os.path.islink(self.symlink_path):
            if os.readlink(self.symlink_path) != old_file:
                if os.path.exists(old_file):
                    os.remove(old_file)
                old_file = os.readlink(self.symlink_path)
            with open(old_file, 'w') as symlink_file:
                try:
                    fcntl.flock(symlink_file, fcntl.LOCK_EX)
                except Exception:
                    sleep(0.1)
                    try:
                        fcntl.flock(symlink_file, fcntl.LOCK_EX)
                    except Exception as e:
                        LOG.error(f"Error lock {old_file}: {str(e)}")
                os.remove(self.symlink_path)
                os.symlink(self._filename, self.symlink_path)
                fcntl.flock(symlink_file, fcntl.LOCK_UN)
            os.remove(old_file)

    @staticmethod
    def metrics_pack(metrics: list):
        text = ''
        if isinstance(metrics, list):
            for v in metrics:
                if isinstance(v, dict):
                    if 'paramname' in v and 'value' in v:
                        text += v['paramname']
                        list_sub_params = []
                        for _k, _v in v.items():
                            if _k != 'paramname' and _k != 'value':
                                list_sub_params.append(f'{_k}="{_v}"')
                        text += f'{{{",".join(list_sub_params)}}} {v["value"]}\n'
        return text.encode('utf-8')

    def write_pack_data(self, metrics: list):
        self.write_data(self.metrics_pack(metrics))
